import ast

from flake8_plugin_utils import Visitor

from flake8_pytest_style.config import Config
from flake8_pytest_style.errors import IncorrectPytestImport


def _is_pytest_or_subpackage(imported_name: str) -> bool:
    return imported_name == 'pytest' or imported_name.startswith('pytest.')


class ImportsVisitor(Visitor[Config]):
    def visit_Import(self, node: ast.Import) -> None:
        for name in node.names:
            if (
                _is_pytest_or_subpackage(name.name)
                and name.asname
                and name.asname != name.name
            ):
                self.error_from_node(IncorrectPytestImport, node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if node.level != 0 or node.module is None:  # relative import
            return
        if _is_pytest_or_subpackage(node.module):
            self.error_from_node(IncorrectPytestImport, node)



import ast
from typing import List, Optional

from flake8_plugin_utils import Visitor, is_none

from flake8_pytest_style.config import Config
from flake8_pytest_style.errors import (
    AssertInExcept,
    RaisesWithMultipleStatements,
    RaisesWithoutException,
    RaisesWithoutMatch,
)
from flake8_pytest_style.utils import (
    get_qualname,
    get_simple_call_args,
    is_raises_call,
    is_raises_with,
)


class RaisesVisitor(Visitor[Config]):
    def __init__(self, config: Optional[Config] = None) -> None:
        super().__init__(config=config)
        self._exception_names: List[str] = []
        self._current_assert: Optional[ast.Assert] = None

    def _check_raises_call(self, node: ast.Call) -> None:
        """
        Checks for violations regarding `pytest.raises` call args (PT010 and PT011).
        """
        args = get_simple_call_args(node)
        exception = args.get_argument('expected_exception', position=0)
        if not exception:
            self.error_from_node(RaisesWithoutException, node)
            return

        exception_name = get_qualname(exception)
        if exception_name not in self.config.raises_require_match_for:
            return
        match = args.get_argument('match')
        if match is None or is_none(match):
            self.error_from_node(RaisesWithoutMatch, node, exception=exception_name)

    def _check_raises_with(self, node: ast.With) -> None:
        """Checks for PT012."""
        body = node.body
        if len(body) != 1 or isinstance(
            body[0],
            (
                ast.If,
                ast.For,
                ast.AsyncFor,
                ast.While,
                ast.With,
                ast.AsyncWith,
                ast.Try,
            ),
        ):
            self.error_from_node(RaisesWithMultipleStatements, node)

    def visit_Call(self, node: ast.Call) -> None:
        if is_raises_call(node):
            self._check_raises_call(node)

    def visit_With(self, node: ast.With) -> None:
        if is_raises_with(node):
            self._check_raises_with(node)

        self.generic_visit(node)

    def visit_ExceptHandler(self, node: ast.ExceptHandler) -> None:
        if node.name:
            self._exception_names.append(node.name)
        try:
            self.generic_visit(node)
        finally:
            if node.name:
                self._exception_names.pop()

    def visit_Assert(self, node: ast.Assert) -> None:
        self._current_assert = node
        try:
            self.visit(node.test)
        finally:
            self._current_assert = None

        if node.msg:
            self.visit(node.msg)

    def visit_Name(self, node: ast.Name) -> None:
        if self._current_assert:
            if node.id in self._exception_names:
                self.error_from_node(AssertInExcept, self._current_assert, name=node.id)



from enum import Enum
from typing import Any, List, NamedTuple, Type


def enum_choices(enum: Type[Enum]) -> List[Any]:
    return [member.value for member in enum]


class ParametrizeNamesType(Enum):
    CSV = 'csv'
    TUPLE = 'tuple'
    LIST = 'list'


class ParametrizeValuesType(Enum):
    TUPLE = 'tuple'
    LIST = 'list'


class ParametrizeValuesRowType(Enum):
    TUPLE = 'tuple'
    LIST = 'list'


class Config(NamedTuple):
    fixture_parentheses: bool
    raises_require_match_for: List[str]
    parametrize_names_type: ParametrizeNamesType
    parametrize_values_type: ParametrizeValuesType
    parametrize_values_row_type: ParametrizeValuesRowType


DEFAULT_CONFIG = Config(
    fixture_parentheses=True,
    raises_require_match_for=[
        'BaseException',
        'Exception',
        'ValueError',
        'IOError',
        'OSError',
        'EnvironmentError',
        'socket.error',
    ],
    parametrize_names_type=ParametrizeNamesType.TUPLE,
    parametrize_values_type=ParametrizeValuesType.LIST,
    parametrize_values_row_type=ParametrizeValuesRowType.TUPLE,
)



import ast
from typing import Dict, NamedTuple, Optional, Tuple, Union

from flake8_plugin_utils.utils import is_false, is_none

AnyFunctionDef = Union[ast.AsyncFunctionDef, ast.FunctionDef]


def get_qualname(node: ast.AST) -> Optional[str]:
    """
    If node represents a chain of attribute accesses, return is qualified name.
    """
    parts = []
    while True:
        if isinstance(node, ast.Name):
            parts.append(node.id)
            break
        if isinstance(node, ast.Attribute):
            parts.append(node.attr)
            node = node.value
        else:
            return None
    return '.'.join(reversed(parts))


class SimpleCallArgs(NamedTuple):
    args: Tuple[ast.AST, ...]
    kwargs: Dict[str, ast.AST]

    def get_argument(
        self, name: str, position: Optional[int] = None
    ) -> Optional[ast.AST]:
        """Get argument by name in kwargs or position in args."""
        kwarg = self.kwargs.get(name)
        if kwarg is not None:
            return kwarg
        if position is not None and len(self.args) > position:
            return self.args[position]
        return None


def get_simple_call_args(node: ast.Call) -> SimpleCallArgs:
    """
    Get call arguments which are specified explicitly (positional and keyword).
    """

    # list of leading non-starred args
    args = []
    for arg in node.args:
        if isinstance(arg, ast.Starred):
            break
        args.append(arg)

    # dict of keyword args
    keywords: Dict[str, ast.AST] = {}
    for keyword in node.keywords:
        if keyword.arg is not None:
            keywords[keyword.arg] = keyword.value

    return SimpleCallArgs(tuple(args), keywords)


def is_parametrize_call(node: ast.Call) -> bool:
    """Checks if given call is to `pytest.mark.parametrize`."""
    return get_qualname(node.func) == 'pytest.mark.parametrize'


def is_raises_call(node: ast.Call) -> bool:
    """Checks if given call is to `pytest.raises`."""
    return get_qualname(node.func) == 'pytest.raises'


def is_fail_call(node: ast.Call) -> bool:
    """Checks if given call is to `pytest.fail`."""
    return get_qualname(node.func) == 'pytest.fail'


def is_raises_with(node: ast.With) -> bool:
    """Checks that a given `with` statement has a `pytest.raises` context."""
    for item in node.items:
        if isinstance(item.context_expr, ast.Call) and is_raises_call(
            item.context_expr
        ):
            return True
    return False


class ParametrizeArgs(NamedTuple):
    names: ast.AST
    values: Optional[ast.AST]
    ids: Optional[ast.AST]


def extract_parametrize_call_args(node: ast.Call) -> Optional[ParametrizeArgs]:
    """Extracts argnames, argvalues and ids from a parametrize call."""
    args = get_simple_call_args(node)

    names_arg = args.get_argument('argnames', 0)
    if names_arg is None:
        return None

    values_arg = args.get_argument('argvalues', 1)
    ids_arg = args.get_argument('ids')
    return ParametrizeArgs(names_arg, values_arg, ids_arg)


def _is_pytest_fixture(node: ast.AST) -> bool:
    """Checks if node is a `pytest.fixture` attribute access."""
    return get_qualname(node) == 'pytest.fixture'


def get_fixture_decorator(node: AnyFunctionDef) -> Union[ast.Call, ast.Attribute, None]:
    """
    Returns a @pytest.fixture decorator applied to given function definition, if any.

    Return value is either:
    * ast.Call, if decorator is written as @pytest.fixture()
    * ast.Attribute, if decorator is written as @pytest.fixture
    * None, if decorator not found
    """
    for decorator in node.decorator_list:
        if (
            isinstance(decorator, ast.Call)
            and isinstance(decorator.func, ast.Attribute)
            and _is_pytest_fixture(decorator.func)
        ):
            return decorator
        if isinstance(decorator, ast.Attribute) and _is_pytest_fixture(decorator):
            return decorator

    return None


def is_empty_string(node: ast.AST) -> bool:
    """
    Checks if the node is a constant empty string.
    """

    # empty string literal
    if isinstance(node, ast.Str) and not node.s:
        return True

    # empty f-string
    if isinstance(node, ast.JoinedStr) and not node.values:
        return True

    return False


def _is_empty_iterable(  # pylint:disable=too-many-return-statements
    node: ast.AST,
) -> bool:
    """
    Checks if the node is a constant empty iterable.
    """

    if is_empty_string(node):
        return True

    # empty list or tuple literal
    if isinstance(node, (ast.List, ast.Tuple)) and not node.elts:
        return True

    # empty dict literal
    if isinstance(node, ast.Dict) and not node.keys:
        return True

    if isinstance(node, ast.Call) and get_qualname(node.func) in (
        'list',
        'set',
        'tuple',
        'dict',
        'frozenset',
    ):
        if not node.args and not node.keywords:  # no args
            return True
        if (
            len(node.args) == 1
            and not node.keywords
            and _is_empty_iterable(node.args[0])
        ):  # single arg, empty iterable
            return True

    return False


def is_falsy_constant(node: ast.AST) -> bool:
    """
    Checks if the node is a constant with a falsy value.
    """

    # None or False constant
    if is_none(node) or is_false(node):
        return True

    # zero literal
    if isinstance(node, ast.Num) and not node.n:
        return True

    return _is_empty_iterable(node)


def is_test_function(node: AnyFunctionDef) -> bool:
    """Checks if the given function is a test function."""
    return node.name.startswith('test_')



from flake8_plugin_utils import Error


class IncorrectFixtureParenthesesStyle(Error):
    code = 'PT001'
    message = 'use @pytest.fixture{expected_parens} over @pytest.fixture{actual_parens}'


class FixturePositionalArgs(Error):
    code = 'PT002'
    message = (
        "configuration for fixture '{name}' specified via positional args, use kwargs"
    )


class ExtraneousScopeFunction(Error):
    code = 'PT003'
    message = "scope='function' is implied in @pytest.fixture()"


class MissingFixtureNameUnderscore(Error):
    code = 'PT004'
    message = "fixture '{name}' does not return anything, add leading underscore"


class IncorrectFixtureNameUnderscore(Error):
    code = 'PT005'
    message = "fixture '{name}' returns a value, remove leading underscore"


class ParametrizeNamesWrongType(Error):
    code = 'PT006'
    message = 'wrong name(s) type in @pytest.mark.parametrize, expected {expected_type}'


class ParametrizeValuesWrongType(Error):
    code = 'PT007'
    message = 'wrong values type in @pytest.mark.parametrize, expected {expected_type}'


class PatchWithLambda(Error):
    code = 'PT008'
    message = 'use return_value= instead of patching with lambda'


class UnittestAssertion(Error):
    code = 'PT009'
    message = "use a regular assert instead of unittest-style '{assertion}'"


class RaisesWithoutException(Error):
    code = 'PT010'
    message = 'set the expected exception in pytest.raises()'


class RaisesWithoutMatch(Error):
    code = 'PT011'
    message = 'set the match parameter in pytest.raises({exception})'


class RaisesWithMultipleStatements(Error):
    code = 'PT012'
    message = 'pytest.raises() block should contain a single simple statement'


class IncorrectPytestImport(Error):
    code = 'PT013'
    message = "found incorrect import of pytest, use simple 'import pytest' instead"


class DuplicateParametrizeTestCases(Error):
    code = 'PT014'
    message = 'found duplicate test cases {indexes} in @pytest.mark.parametrize'


class AssertAlwaysFalse(Error):
    code = 'PT015'
    message = 'assertion always fails, replace with pytest.fail()'


class FailWithoutMessage(Error):
    code = 'PT016'
    message = 'no message passed to pytest.fail()'


class AssertInExcept(Error):
    code = 'PT017'
    message = (
        'found assertion on exception {name} in except block,'
        ' use pytest.raises() instead'
    )


class CompositeAssertion(Error):
    code = 'PT018'
    message = 'assertion should be broken down into multiple parts'


class FixtureParamWithoutValue(Error):
    code = 'PT019'
    message = (
        'fixture {name} without value is injected as parameter,'
        ' use @pytest.mark.usefixtures instead'
    )



import ast

from flake8_plugin_utils import Visitor

from flake8_pytest_style.config import Config
from flake8_pytest_style.errors import CompositeAssertion, UnittestAssertion

_UNITTEST_ASSERT_NAMES = (
    'assertAlmostEqual',
    'assertAlmostEquals',
    'assertDictEqual',
    'assertEqual',
    'assertEquals',
    'assertFalse',
    'assertGreater',
    'assertGreaterEqual',
    'assertIn',
    'assertIs',
    'assertIsInstance',
    'assertIsNone',
    'assertIsNot',
    'assertIsNotNone',
    'assertItemsEqual',
    'assertLess',
    'assertLessEqual',
    'assertMultiLineEqual',
    'assertNotAlmostEqual',
    'assertNotAlmostEquals',
    'assertNotContains',
    'assertNotEqual',
    'assertNotEquals',
    'assertNotIn',
    'assertNotIsInstance',
    'assertNotRegexpMatches',
    'assertRaises',
    'assertRaisesMessage',
    'assertRaisesRegexp',
    'assertRegexpMatches',
    'assertSetEqual',
    'assertTrue',
    'assert_',
)


class UnittestAssertionVisitor(Visitor[Config]):
    def visit_Call(self, node: ast.Call) -> None:
        if (
            isinstance(node.func, ast.Attribute)
            and node.func.attr in _UNITTEST_ASSERT_NAMES
        ):
            self.error_from_node(UnittestAssertion, node, assertion=node.func.attr)


class AssertionVisitor(Visitor[Config]):
    def _is_composite_condition(self, expr: ast.AST) -> bool:
        # e.g. `a and b`
        if isinstance(expr, ast.BoolOp) and isinstance(expr.op, ast.And):
            return True

        # e.g. `not (a or b)`
        if (
            isinstance(expr, ast.UnaryOp)
            and isinstance(expr.op, ast.Not)
            and isinstance(expr.operand, ast.BoolOp)
            and isinstance(expr.operand.op, ast.Or)
        ):
            return True

        return False

    def visit_Assert(self, node: ast.Assert) -> None:
        if self._is_composite_condition(node.test):
            self.error_from_node(CompositeAssertion, node)



import ast

from flake8_plugin_utils import Visitor

from flake8_pytest_style.config import Config
from flake8_pytest_style.errors import AssertAlwaysFalse, FailWithoutMessage
from flake8_pytest_style.utils import (
    get_simple_call_args,
    is_empty_string,
    is_fail_call,
    is_falsy_constant,
)


class FailVisitor(Visitor[Config]):
    def _check_fail_call(self, node: ast.Call) -> None:
        """Checks for PT016."""
        args = get_simple_call_args(node)
        msg_argument = args.get_argument('msg', 0)
        if not msg_argument or is_empty_string(msg_argument):
            self.error_from_node(FailWithoutMessage, node)

    def visit_Assert(self, node: ast.Assert) -> None:
        """Checks for PT015."""
        if is_falsy_constant(node.test):
            self.error_from_node(AssertAlwaysFalse, node)

    def visit_Call(self, node: ast.Call) -> None:
        if is_fail_call(node):
            self._check_fail_call(node)



import argparse
from typing import List

from flake8.options.manager import OptionManager
from flake8_plugin_utils import Plugin

from .config import (
    DEFAULT_CONFIG,
    Config,
    ParametrizeNamesType,
    ParametrizeValuesRowType,
    ParametrizeValuesType,
    enum_choices,
)
from .visitors import (
    AssertionVisitor,
    FailVisitor,
    FixturesVisitor,
    ImportsVisitor,
    ParametrizeVisitor,
    PatchVisitor,
    RaisesVisitor,
    UnittestAssertionVisitor,
)

__version__ = '1.1.0'


class PytestStylePlugin(Plugin[Config]):
    name = 'flake8-pytest-style'
    version = __version__
    visitors = [
        AssertionVisitor,
        FailVisitor,
        FixturesVisitor,
        ImportsVisitor,
        PatchVisitor,
        ParametrizeVisitor,
        RaisesVisitor,
        UnittestAssertionVisitor,
    ]

    @classmethod
    def add_options(cls, option_manager: OptionManager) -> None:
        option_manager.add_option(
            '--pytest-fixture-no-parentheses',
            action='store_true',
            parse_from_config=True,
            default=not DEFAULT_CONFIG.fixture_parentheses,
            help='Omit parentheses for @pytest.fixture decorators'
            ' without parameters. (Default: %default)',
        )
        option_manager.add_option(
            '--pytest-raises-require-match-for',
            comma_separated_list=True,
            parse_from_config=True,
            default=DEFAULT_CONFIG.raises_require_match_for,
            help='List of exceptions for which flake8-pytest-style requires'
            ' a match= argument in pytest.raises(). (Default: %default)',
        )
        option_manager.add_option(
            '--pytest-parametrize-names-type',
            choices=enum_choices(ParametrizeNamesType),
            parse_from_config=True,
            default=DEFAULT_CONFIG.parametrize_names_type.value,
            help='Preferred type for multiple parameter names in'
            ' @pytest.mark.parametrize. (Default: %default)',
        )
        option_manager.add_option(
            '--pytest-parametrize-values-type',
            choices=enum_choices(ParametrizeValuesType),
            parse_from_config=True,
            default=DEFAULT_CONFIG.parametrize_values_type.value,
            help='Preferred type for values in @pytest.mark.parametrize.'
            ' (Default: %default)',
        )
        option_manager.add_option(
            '--pytest-parametrize-values-row-type',
            choices=enum_choices(ParametrizeValuesRowType),
            parse_from_config=True,
            default=DEFAULT_CONFIG.parametrize_values_row_type.value,
            help='Preferred type for each row in @pytest.mark.parametrize'
            ' in case of multiple parameters. (Default: %default)',
        )

    @classmethod
    def parse_options_to_config(  # pylint: disable=unused-argument
        cls, option_manager: OptionManager, options: argparse.Namespace, args: List[str]
    ) -> Config:
        return Config(
            fixture_parentheses=not options.pytest_fixture_no_parentheses,
            raises_require_match_for=options.pytest_raises_require_match_for,
            parametrize_names_type=ParametrizeNamesType(
                options.pytest_parametrize_names_type
            ),
            parametrize_values_type=ParametrizeValuesType(
                options.pytest_parametrize_values_type
            ),
            parametrize_values_row_type=ParametrizeValuesRowType(
                options.pytest_parametrize_values_row_type
            ),
        )



from .assertion import AssertionVisitor, UnittestAssertionVisitor
from .fail import FailVisitor
from .fixtures import FixturesVisitor
from .imports import ImportsVisitor
from .parametrize import ParametrizeVisitor
from .patch import PatchVisitor
from .raises import RaisesVisitor

__all__ = (
    'AssertionVisitor',
    'FailVisitor',
    'FixturesVisitor',
    'ImportsVisitor',
    'ParametrizeVisitor',
    'PatchVisitor',
    'RaisesVisitor',
    'UnittestAssertionVisitor',
)



import ast
from typing import Union

from flake8_plugin_utils import Visitor

from flake8_pytest_style.config import Config
from flake8_pytest_style.errors import (
    ExtraneousScopeFunction,
    FixtureParamWithoutValue,
    FixturePositionalArgs,
    IncorrectFixtureNameUnderscore,
    IncorrectFixtureParenthesesStyle,
    MissingFixtureNameUnderscore,
)
from flake8_pytest_style.utils import (
    AnyFunctionDef,
    get_fixture_decorator,
    is_test_function,
)


class FixturesVisitor(Visitor[Config]):
    def _check_fixture_decorator(
        self,
        fixture_decorator: Union[ast.Call, ast.Attribute],
        fixture_func: AnyFunctionDef,
    ) -> None:
        """Checks for PT001, PT002, PT003."""
        if not isinstance(fixture_decorator, ast.Call):
            if self.config.fixture_parentheses:
                self.error_from_node(
                    IncorrectFixtureParenthesesStyle,
                    fixture_decorator,
                    expected_parens='()',
                    actual_parens='',
                )
            return

        if (
            not self.config.fixture_parentheses
            and not fixture_decorator.args
            and not fixture_decorator.keywords
        ):
            self.error_from_node(
                IncorrectFixtureParenthesesStyle,
                fixture_decorator,
                expected_parens='',
                actual_parens='()',
            )

        if fixture_decorator.args:
            self.error_from_node(
                FixturePositionalArgs, fixture_decorator, name=fixture_func.name
            )

        for keyword in fixture_decorator.keywords:
            if (
                keyword.arg == 'scope'
                and isinstance(keyword.value, ast.Str)
                and keyword.value.s == 'function'
            ):
                self.error_from_node(
                    ExtraneousScopeFunction, fixture_decorator, name=fixture_func.name
                )

    def _check_fixture_returns(self, node: AnyFunctionDef) -> None:
        """Checks for PT004, PT005."""
        has_return_with_value = False
        for child in ast.walk(node):
            if isinstance(child, (ast.Return, ast.Yield)) and child.value is not None:
                has_return_with_value = True
                break

        if has_return_with_value and node.name.startswith('_'):
            self.error_from_node(IncorrectFixtureNameUnderscore, node, name=node.name)
        elif not has_return_with_value and not node.name.startswith('_'):
            self.error_from_node(MissingFixtureNameUnderscore, node, name=node.name)

    def _check_test_function_args(self, node: AnyFunctionDef) -> None:
        """Checks for PT019."""
        # intentionally not looking at posonlyargs because pytest passes everything
        # as kwargs, so declaring fixture args as positional-only will fail anyway
        for arg in node.args.args + node.args.kwonlyargs:
            if arg.arg.startswith('_'):
                self.error_from_node(FixtureParamWithoutValue, node, name=arg.arg)

    def visit_FunctionDef(self, node: AnyFunctionDef) -> None:
        fixture_decorator = get_fixture_decorator(node)
        if fixture_decorator:
            self._check_fixture_decorator(fixture_decorator, node)
            self._check_fixture_returns(node)

        if is_test_function(node):
            self._check_test_function_args(node)

        self.generic_visit(node)

    visit_AsyncFunctionDef = visit_FunctionDef  # noqa: N815



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flake8_pytest_style', 'flake8_pytest_style.visitors']

package_data = \
{'': ['*']}

install_requires = \
['flake8-plugin-utils>=1.3.0,<2.0.0']

entry_points = \
{'flake8.extension': ['PT = flake8_pytest_style.plugin:PytestStylePlugin']}

setup_kwargs = {
    'name': 'flake8-pytest-style',
    'version': '1.1.0',
    'description': 'A flake8 plugin checking common style issues or inconsistencies with pytest-based tests.',
    'long_description': "# flake8-pytest-style\n\n[![pypi](https://badge.fury.io/py/flake8-pytest-style.svg)](https://pypi.org/project/flake8-pytest-style)\n[![Python: 3.6+](https://img.shields.io/badge/Python-3.6+-blue.svg)](https://pypi.org/project/flake8-pytest-style)\n[![Downloads](https://img.shields.io/pypi/dm/flake8-pytest-style.svg)](https://pypistats.org/packages/flake8-pytest-style)\n[![Build Status](https://travis-ci.com/m-burst/flake8-pytest-style.svg?branch=master)](https://travis-ci.com/m-burst/flake8-pytest-style)\n[![Code coverage](https://codecov.io/gh/m-burst/flake8-pytest-style/branch/master/graph/badge.svg)](https://codecov.io/gh/m-burst/flake8-pytest-style)\n[![License: MIT](https://img.shields.io/badge/License-MIT-green.svg)](https://en.wikipedia.org/wiki/MIT_License)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n## Description\n\nA `flake8` plugin checking common style issues or inconsistencies with `pytest`-based tests.\n\nCurrently the following errors are reported:\n\n* `PT001 use @pytest.fixture() over @pytest.fixture`  \nThe preferred style is controlled by the configuration option\n`pytest-fixture-no-parentheses`\n\n* `PT002 configuration for fixture '{name}' specified via positional args, use kwargs`  \ne.g. `@pytest.fixture(scope='module')` is OK, and `@pytest.fixture('module')` is an error\n\n* `PT003 scope='function' is implied in @pytest.fixture()`  \ne.g. `@pytest.fixture(scope='function')` should be replaced with `@pytest.fixture()`\n\n* `PT004 fixture '{name}' does not return anything, add leading underscore`\n\n* `PT005 fixture '{name}' returns a value, remove leading underscore`\n\n* `PT006 wrong name(s) type in @pytest.mark.parametrize, expected {expected_type}`  \nFor a single name the expected type is always a plain string.\nFor multiple names the expected type is controlled by the configuration\nvariable `pytest-parametrize-names-type`.\n\n* `PT007 wrong values type in @pytest.mark.parametrize, expected {expected_type}`\nThe expected type of the list of rows is controlled by the configuration\nvariable `pytest-parametrize-values-type`.\nThe expected type of each row in case of multiple arguments is controlled by\nthe configuration variable `pytest-parametrize-values-row-type`.\n\n* `PT008 use return_value= instead of patching with lambda`  \ne.g. `mocker.patch('target', return_value=7)` is OK,\nand `mocker.patch('target', lambda *args: 7)` is an error\n\n* `PT009 use a regular assert instead of unittest-style '{assertion}'`\n\n* `PT010 set the expected exception in pytest.raises()`\n\n* `PT011 set the match parameter in pytest.raises({exception})`  \ne.g. `pytest.raises(ValueError, match='exception text')` is OK,\nand `pytest.raises(ValueError)` is an error.  \nThe exception list is controlled by the configuration option \n`pytest-raises-require-match-for`.\n\n* `PT012 pytest.raises() block should contain a single simple statement`  \nthis forbids multiple statements and control flow structures within\n`pytest.raises()` blocks.\n\n* `PT013 found incorrect import of pytest, use simple 'import pytest' instead`\n\n* `PT014 found duplicate test cases {indexes} in @pytest.mark.parametrize`\n\n* `PT015 assertion always fails, replace with pytest.fail()`  \ne.g. `pytest.fail()` should be preferred to `assert False` or `assert 0`\n\n* `PT016 no message passed to pytest.fail()`\n\n* `PT017 found assertion on exception {name} in except block, use pytest.raises() instead`  \ne.g.:\n  ```python\n  try:\n      1 / 0\n  except ZeroDivisionError as e:\n      assert e.args\n  ```\n  is an error and should be replaced with\n  ```python\n  with pytest.raises(ZeroDivisionError) as e:\n      1 / 0\n  assert e.value.args\n  ```\n  \n* `PT018 assertion should be broken down into multiple parts`  \ne.g. `assert something and something_else` should be rewritten as separate\n`assert something` and `assert something_else`,\nand `assert not (something or something_else)` should be rewritten as\n`assert not something` and `assert not something_else`\n\n* `PT019 fixture {name} without value is injected as parameter, use @pytest.mark.usefixtures instead`\n\n## Installation\n\n    pip install flake8-pytest-style\n    \n## Configuration\n\nThe plugin has the following configuration options:\n\n* `pytest-fixture-no-parentheses`  \nBoolean flag specifying whether `@pytest.fixture()` without parameters\nshould have parentheses, as checked by `PT001`.  \nIf the option is set to false (the default), `@pytest.fixture()` is valid\nand `@pytest.fixture` is an error.  \nIf set to true, `@pytest.fixture` is valid and `@pytest.fixture()` is\nan error.\n\n* `pytest-parametrize-names-type`  \nExpected type for multiple argument names in `@pytest.mark.parametrize`,\nas checked by `PT006`. The following values are supported:\n  * `csv` &mdash; a comma-separated list, e.g. `@pytest.mark.parametrize('name1,name2', ...)`\n  * `tuple` (default) &mdash; e.g. `@pytest.mark.parametrize(('name1', 'name2'), ...)`\n  * `list` &mdash; e.g. `@pytest.mark.parametrize(['name1', 'name2'], ...)`\n  \n* `pytest-parametrize-values-type`  \nExpected type for the list of values rows in `@pytest.mark.parametrize`,\nas checked by `PT007`. The following values are supported:\n  * `tuple` &mdash; e.g. `@pytest.mark.parametrize('name', (1, 2, 3))`\n  * `list` (default) &mdash; e.g. `@pytest.mark.parametrize('name', [1, 2, 3])`\n\n* `pytest-parametrize-values-row-type`  \nExpected type for each row of values in `@pytest.mark.parametrize` in case of\nmultiple parameters, as checked by `PT007`. The following values are supported:\n  * `tuple` (default) &mdash; e.g. `@pytest.mark.parametrize(('name1', 'name2'), [(1, 2), (3, 4)])`\n  * `list` &mdash; e.g. `@pytest.mark.parametrize(('name1', 'name2'), [[1, 2], [3, 4]])`\n\n* `pytest-raises-require-match-for`  \nComma-separated list of exception names that require a `match=` parameter\nin a `pytest.raises()` call, as checked by `PT011`. By default the list\ncontains the following exceptions:\n  * `BaseException`, `Exception`\n  * `ValueError`\n  * `OSError`, `IOError`, `EnvironmentError`, `socket.error`\n\n## For developers\n\n### Install deps and setup pre-commit hook\n\n    make init\n\n### Run linters, autoformat, tests etc.\n\n    make format lint test\n\n### Bump new version\n\n    make bump_major\n    make bump_minor\n    make bump_patch\n\n## License\n\nMIT\n\n## Change Log\n\n**Unreleased**\n\n...\n\n**1.1.0 - 2020-04-14**\n\n* add `PT015` (checks for `assert False`)\n* add `PT016` (checks for `pytest.fail()` without message)\n* add `PT017` (checks for assertions on exceptions in `except` blocks)\n* add `PT018` (checks for composite assertions)\n* add `PT019` (checks for fixtures without value injected as parameters)\n\n**1.0.0 - 2020-03-26**\n\n* add `PT014` (checks for duplicate test cases in `@pytest.mark.parametrize`)\n\n**0.6.0 - 2020-03-21**\n\n* add configuration option `pytest-parametrize-names-type` for `PT006`\n* add configuration options `pytest-parametrize-values-type` and\n`pytest-parametrize-values-row-type` for `PT007`\n\n**0.5.0 - 2020-03-09**\n\n* add configuration option `pytest-fixture-no-parentheses` for `PT001`\n* add `PT013` (checks for `from`-imports from `pytest`)\n\n**0.4.0 - 2020-03-09**\n\n* add `PT012` (checks for multiple statements in `with pytest.raises()`)\n\n**0.3.1 - 2020-03-09**\n\n* fix default value of `pytest-raises-require-match-for` config option\n\n**0.3.0 - 2020-03-09**\n\n* add `PT010` and `PT011` (checks for `pytest.raises` parameters)\n\n**0.2.0 - 2020-03-01**\n\n* add `PT009` (ported from [flake8-pytest](https://github.com/vikingco/flake8-pytest))\n\n**0.1.3 - 2019-05-24**\n\n* add `yield` fixtures support\n* fix changelog entry for 0.1.2\n\n**0.1.2 - 2019-05-23**\n\n* fix parametrize checkers not working in decorators\n\n**0.1.1 - 2019-05-23**\n\n* update PyPI description\n\n**0.1.0 - 2019-05-23**\n\n* initial\n",
    'author': 'Mikhail Burshteyn',
    'author_email': 'mdburshteyn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pypi.org/project/flake8-pytest-style',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import ast
import itertools
from typing import Optional

from flake8_plugin_utils import Visitor, check_equivalent_nodes

from ..config import (
    Config,
    ParametrizeNamesType,
    ParametrizeValuesRowType,
    ParametrizeValuesType,
)
from ..errors import (
    DuplicateParametrizeTestCases,
    ParametrizeNamesWrongType,
    ParametrizeValuesWrongType,
)
from ..utils import extract_parametrize_call_args, is_parametrize_call


class ParametrizeVisitor(Visitor[Config]):
    def _check_parametrize_names(
        self, node: ast.Call, names: ast.AST
    ) -> Optional[bool]:
        """
        Handles names in parametrize, checks for PT006.

        Returns a flag indicating whether parametrize has multiple names,
        or None if we can't tell.
        """

        multiple_names: Optional[bool] = None
        found_type: Optional[ParametrizeNamesType] = None
        if isinstance(names, ast.Str):
            if ',' in names.s:
                found_type = ParametrizeNamesType.CSV
                multiple_names = True
            else:
                multiple_names = False
        elif isinstance(names, (ast.List, ast.Tuple)):
            multiple_names = len(names.elts) > 1
            if not multiple_names:
                self.error_from_node(
                    ParametrizeNamesWrongType, node, expected_type='string'
                )
            elif isinstance(names, ast.Tuple):
                found_type = ParametrizeNamesType.TUPLE
            else:
                found_type = ParametrizeNamesType.LIST
        if multiple_names and found_type != self.config.parametrize_names_type:
            self.error_from_node(
                ParametrizeNamesWrongType,
                node,
                expected_type=self.config.parametrize_names_type.value,
            )
        return multiple_names

    def _get_expected_values_type_str(self, multiple_names: Optional[bool]) -> str:
        if multiple_names:
            return (
                f'{self.config.parametrize_values_type.value}'
                f' of {self.config.parametrize_values_row_type.value}s'
            )
        return self.config.parametrize_values_type.value

    def _check_parametrize_values(
        self, node: ast.Call, values: Optional[ast.AST], multiple_names: Optional[bool]
    ) -> None:
        """Checks for PT007."""
        expected_type_str = self._get_expected_values_type_str(multiple_names)

        if isinstance(values, ast.List):
            top_level_type = ParametrizeValuesType.LIST
        elif isinstance(values, ast.Tuple):
            top_level_type = ParametrizeValuesType.TUPLE
        else:
            return

        if top_level_type != self.config.parametrize_values_type:
            self.error_from_node(
                ParametrizeValuesWrongType, node, expected_type=expected_type_str
            )
            return

        if multiple_names:
            for element in values.elts:
                found_row_type: Optional[ParametrizeValuesRowType] = None
                if isinstance(element, ast.List):
                    found_row_type = ParametrizeValuesRowType.LIST
                elif isinstance(element, ast.Tuple):
                    found_row_type = ParametrizeValuesRowType.TUPLE
                if (
                    found_row_type
                    and found_row_type != self.config.parametrize_values_row_type
                ):
                    self.error_from_node(
                        ParametrizeValuesWrongType,
                        node,
                        expected_type=expected_type_str,
                    )
                    break

    def _check_parametrize_duplicates(
        self, node: ast.AST, values: Optional[ast.AST]
    ) -> None:
        """Checks for PT014."""
        if not isinstance(values, (ast.List, ast.Tuple, ast.Set)):
            return

        for (i, element1), (j, element2) in itertools.combinations(
            enumerate(values.elts, start=1), 2
        ):
            if check_equivalent_nodes(element1, element2):
                self.error_from_node(
                    DuplicateParametrizeTestCases, node, indexes=(i, j)
                )

    def _check_parametrize_call(self, node: ast.Call) -> None:
        """Checks for all violations regarding `pytest.mark.parametrize` calls."""
        args = extract_parametrize_call_args(node)
        if not args:
            return

        multiple_names = self._check_parametrize_names(node, args.names)

        self._check_parametrize_values(node, args.values, multiple_names)

        self._check_parametrize_duplicates(node, args.values)

    def visit_Call(self, node: ast.Call) -> None:
        if is_parametrize_call(node):
            self._check_parametrize_call(node)



import ast

from flake8_plugin_utils import Visitor

from flake8_pytest_style.config import Config
from flake8_pytest_style.errors import PatchWithLambda
from flake8_pytest_style.utils import get_qualname, get_simple_call_args

_PATCH_NAMES = ('mocker.patch', 'mock.patch', 'unittest.mock.patch', 'patch')
_PATCH_OBJECT_NAMES = tuple(f'{name}.object' for name in _PATCH_NAMES)


class PatchVisitor(Visitor[Config]):
    def _check_patch_call(self, node: ast.Call, new_arg_number: int) -> None:
        """
        Checks for PT008.

        :param node: patch call node
        :param new_arg_number: number of `new` positional argument of patch func
        """
        args = get_simple_call_args(node)
        if args.get_argument('return_value') is not None:
            return

        new_arg = args.get_argument('new', new_arg_number)
        if not isinstance(new_arg, ast.Lambda):
            return

        lambda_argnames = {
            arg.arg for arg in new_arg.args.args + new_arg.args.kwonlyargs
        }
        if new_arg.args.vararg:
            lambda_argnames.add(new_arg.args.vararg.arg)
        if new_arg.args.kwarg:
            lambda_argnames.add(new_arg.args.kwarg.arg)

        for child_node in ast.walk(new_arg.body):
            if isinstance(child_node, ast.Name) and child_node.id in lambda_argnames:
                break
        else:
            self.error_from_node(PatchWithLambda, node)

    def visit_Call(self, node: ast.Call) -> None:
        if get_qualname(node.func) in _PATCH_NAMES:
            # attributes are (target, new, ...)
            self._check_patch_call(node, 1)

        if get_qualname(node.func) in _PATCH_OBJECT_NAMES:
            # attributes are (target, attribute, new, ...)
            self._check_patch_call(node, 2)
